@@ -15,9 +15,11 @@ def get_K_identity(K, p):
     return res
 
 def sparsity(S):
+    """off-diagonal ratio of nonzero entries in S"""
+    assert len(S.shape) == 2
     (p,p) = S.shape
-    A = adjacency_matrix(S)
-    s = A.sum()/(p**2-p)
+    off_nnz = np.count_nonzero(S) - p
+    s = off_nnz/(p**2-p)
     return s
 
 def mean_sparsity(S):
